require 'lx'

#===============================================================================
# Talius
#

# A +Talius+ object represents a full CSS selector.
class Talius
	# version '0.6'
	VERSION = '0.6'
	
	# An array of rule objects.
	attr_reader :rules
	
	#---------------------------------------------------------------------------
	# initialize
	#
	
	# Initialize a new Talius object. Takes a raw CSS selector string as the
	# single param.
	#
	#  raw = 'a[href]'
	#  selector = Talius.new(raw)
	def initialize(p_raw)
		@raw = p_raw
		@rules = []
		@slashes = {}
		@quotes = {}
		@norm = nil
		
		# normalize selector string
		unslash()
		unquote()
		@norm.lx.collapse!
		
		# split into rules
		@norm.split(/\s*,\s*/mu).each do |rule_str|
			@rules.push Talius::Rule.new(self, rule_str)
		end
	end
	#
	# initialize
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# raw
	#
	
	# The original string used to the create the selector object.
	#
	#  raw = 'a[href]'
	#  selector = Talius.new(raw)
	#  selector.raw # => "a[href]"
	attr_reader :raw
	
	#
	# raw
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# denormalize
	#
	
	# Used by Talius::Rule to rebuild a string that has placeholders in it. This
	# method is not useful for any purpose beyond that.
	def denormalize(raw)
		return reslash(requote(raw))
	end
	#
	# denormalize
	#---------------------------------------------------------------------------
	
	
	# private
	private
	
	
	#---------------------------------------------------------------------------
	# requote
	#
	def requote(raw)
		return @quotes[raw] || raw
	end
	#
	# requote
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# reslash
	#
	def reslash(rv)
		# $tm.hrm
		
		# loop through slashed strings
		@slashes.each do |pattern, val|
			rv = rv.sub(/#{pattern}/mu, val)
		end
		
		# return
		return rv
	end
	#
	# reslash
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# unquote
	#
	def unquote()
		rv = []
		
		# split into slashed and unslashed strings
		@norm.split(/("[^"]*")/mu).each do |token|
			if token.match(/\A".*"\z/mu)
				content = token
				content = content.sub(/\A"/mu, '')
				content = content.sub(/"\z/mu, '')
				ph = LX.randstr()
				@quotes[ph] = content
				rv.push ph
			else
				rv.push token
			end
		end
		
		# return new string
		@norm = rv.join('')
	end
	#
	# unquote
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# unslash
	#
	def unslash()
		rv = []
		
		# split into slashed and unslashed strings
		@raw.split(/(\\.)/mu).each do |token|
			if token.match(/\A\\.\z/mu)
				ph = LX.randstr()
				@slashes[ph] = token[1]
				rv.push ph
			else
				rv.push token
			end
		end
		
		# return new string
		@norm = rv.join('')
	end
	#
	# unslash
	#---------------------------------------------------------------------------
end
#
# Talius
#===============================================================================


#===============================================================================
# Talius::Rule
#

# Talius::Rule object represents one set of directives, e.g. tags, classes, etc.
class Talius::Rule
	#---------------------------------------------------------------------------
	# special characters
	#
	@@NESTING_CHARS = %w{> + ~ s}
	@@NESTING_RX = @@NESTING_CHARS.map{|el| '\\' + el}.join('')
	@@SUBRULE_CHARS = %w{. [ ] : #}
	@@SUBRULE_RX = @@SUBRULE_CHARS.map{|el| '\\' + el}.join('')
	@@META_RX = @@NESTING_RX + @@SUBRULE_RX
	#
	# special characters
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# initialize
	# Lexical parsing experts will probably recoil in horror at how I parse a
	# selector rule. Whatever.
	#
	
	# Initializes a Talius::Rule object. The first param is the Talius object
	# with which the rule will be associated. The second param is the raw string
	# that describes the rule.
	#
	#  raw = 'section, div'
	#  selector = Talius.new(raw)
	#  selector.rules[0] # => {"tags"=>{"section"=>{"name"=>"section"}}}
	def initialize(p_selector, raw)
		@selector = p_selector
		@classes = {}
		@tags = {}
		@ids = {}
		@pelements = {}
		@pclasses = {}
		@atts = {}
		
		# remove spaces before and after parens
		raw = raw.gsub(/\s*([\(\)])\s*/mu, '\1')
		
		# split into subrules
		pieces = raw.split(/(\[[^\[\]]*\])|(\:\:[^#{@@META_RX}]+)|([#{@@NESTING_RX}])|([#{@@META_RX}][^#{@@META_RX}]*)/mu)
		pieces = pieces.grep(/./mu)
		
		# build subrules
		pieces.each do |piece|
			# I haven't yet implemented nested rules, so at this point we should
			# not have any pieces that consist of just spaces
			if @@NESTING_CHARS.include?(piece)
				raise 'not-yet-implemented-nested-elements: ' + piece
			end
			
			# if piece consists of a single meta character, that's invalid
			if @@SUBRULE_CHARS.include?(piece)
				raise 'meta-character-without-value: ' + piece
			end
			
			# class
			if piece.match(/\A\..+/mu)
				@classes[piece.sub(/\A\./mu, '')] = true
			
			# pseudo element ::
			elsif piece.match(/\A\:\:/mu)
				pe = piece.sub(/\A\:\:/mu, '')
				pseudo @pelements, pe
				
			# pseudo class :
			elsif piece.match(/\A\:/mu)
				pc = piece.sub(/\A\:/mu, '')
				pseudo @pclasses, pc
				
			# attribute
			elsif piece.match(/\A\[.*\]\z/mu)
				full = piece
				full = full.sub(/\A\s*\[\s*/mu, '')
				full = full.sub(/\s*\]\s*\z/mu, '')
				nv = full.split(/\s*\=\s*/mu, 2)
				
				if nv[0]
					att = Talius::Node::Att.new(@selector.denormalize(nv[0]))
					@atts[att.full_name] = att
					
					if nv[1]
						att.value = @selector.denormalize(nv[1])
					end
				end
			
			# id
			elsif piece.match(/\A\#/mu)
				id = piece
				id = id.sub(/\A\#/mu, '')
				@ids[id] = true
			
			# else tag
			else
				tag = Talius::Node::Tag.new(@selector.denormalize(piece))
				@tags[tag.full_name] = tag
			end
		end
	end
	#
	# initialize
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# selector
	#
	
	# The +Talius+ object that this rule is associated with.
	attr_reader :selector
	
	#
	# selector
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# classes
	#
	
	# A hash of the classes described in the selector. The hash keys are the
	# names of the classes. The values are always true.
	#
	#  raw = 'section.ready.overview'
	#  selector = Talius.new(raw)
	#  selector.rules[0].classes # => {"ready"=>true, "overview"=>true}
	attr_reader :classes
	
	#
	# classes
	#---------------------------------------------------------------------------
	

	#---------------------------------------------------------------------------
	# tags
	#
	
	# A hash of the tags described in the selector. The hash keys are the names
	# if the tags. The values are Talius::Node::Tag objects. Currently, Talius
	# can only parse a single tag from a selector rule, so for now this hash
	# will always have a single element.
	#
	#  raw = 'section'
	#  selector = Talius.new(raw)
	#  selector.rules[0].tags # => {"section"=>{"name"=>"section"}}
	attr_reader :tags
	
	#
	# tags
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# ids
	#
	
	# A hash of elements defined by ID. The hash keys are the id's, the
	# values are always true
	#
	#  raw = '#overview'
	#  selector = Talius.new(raw)
	#  rule = selector.rules[0]
	#  rule.ids # => {"overview"=>true}
	
	attr_reader :ids
	
	#
	# ids
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# pelements
	#
	
	# A hash of pseudo-elements such as +::first-line+. The hash keys are the
	# names of the pseudo-elements. The values are either the value in the
	# parens after the name of the pseudo-element, or, if there is no such
	# value, +nil+.
	#
	#  raw = 'section::first-line'
	#  selector = Talius.new(raw)
	#  selector.rules[0].pelements # => {"first-line"=>nil}
	#
	#  raw = 'section::nth-line(9)'
	#  selector = Talius.new(raw)
	#  selector.rules[0].pelements # => {"nth-line"=>"9"}
	
	attr_reader :pelements
	
	#
	# pelements
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# pclasses
	#
	
	# A hash of pseudo-classes such as +:visited+. The hash keys are the
	# names of the pseudo-elements. The values are either the value in the
	# parens after the name of the pseudo-element, or, if there is no such
	# value, +nil+.
	#
	#  raw = 'a:visited'
	#  selector = Talius.new(raw)
	#  selector.rules[0].pclasses # => {"visited"=>nil}
	#
	#  raw = 'ul:nth-child(9)'
	#  selector = Talius.new(raw)
	#  selector.rules[0].pclasses # => {"nth-child"=>"9"}
	
	attr_reader :pclasses
	
	#
	# pclasses
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# atts
	#
	
	# Hash of attributes. The hash keys are the names of the attributes. The
	# values are Talius::Node::Att objects.
	#
	#  raw = 'a[href][rel=parent]'
	#  selector = Talius.new(raw)
	#  selector.rules[0].atts # => {"href"=>{"name"=>"href"}, "rel"=>{"name"=>"rel", "value"=>"parent"}}
	
	attr_reader :atts
	
	#
	# atts
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# to_h
	#
	
	# Returns a hash representation of the rule.
	def to_h
		# $tm.hrm
		rv = {}
		
		# tags
		if @tags.any?
			rv['tags'] = {}.tap do |tags|
				@tags.each do |key, node|
					tags[node.full_name] = node.to_h
				end
			end
		end
		
		# ids
		if @ids.any?
			rv['ids'] = {}.tap do |ids|
				@ids.each do |key, val|
					ids[key] = val
				end
			end
		end
		
		# atts
		if @atts.any?
			rv['atts'] = {}.tap do |atts|
				@atts.each do |key, node|
					atts[node.full_name] = node.to_h
				end
			end
		end
		
		# classes
		if @classes.any?
			rv['classes'] = {}.tap do |classes|
				@classes.each do |key, val|
					classes[key] = val
				end
			end
		end
		
		# pclasses
		if @pclasses.any?
			rv['pclasses'] = {}.tap do |pclasses|
				@pclasses.each do |key, val|
					pclasses[key] = val
				end
			end
		end
		
		# pelements
		if @pelements.any?
			rv['pelements'] = {}.tap do |pelements|
				@pelements.each do |key, val|
					pelements[key] = val
				end
			end
		end
		
		# return
		return rv
	end
	#
	# to_h
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# inspect
	#
	
	# Returns a stringification of the results of to_h.
	def inspect
		return to_h.to_s
	end
	#
	# inspect
	#---------------------------------------------------------------------------
	
	
	# private
	private
	
	
	#---------------------------------------------------------------------------
	# pseudo
	#
	def pseudo(hsh, raw)
		# get value in parens
		if raw.match(/\(/mu)
			val = raw
			raw = raw.sub(/\(.*/mu, '')
			val = val.sub(/\A.*\(/mu, '')
			val = val.sub(/\)\z/mu, '')
			hsh[@selector.denormalize(raw).downcase] = @selector.denormalize(val)
		else
			hsh[@selector.denormalize(raw).downcase] = nil
		end
	end
	#
	# pseudo
	#---------------------------------------------------------------------------
end
#
# Talius::Rule
#===============================================================================


#===============================================================================
# Talius::Node
#

# Base class for Talius::Node::Tag and Talius::Node::Att.
class Talius::Node
	#---------------------------------------------------------------------------
	# initialize
	#
	
	# Initializes a new node. Accepts the string from the selector. Generally,
	# you won't have to instantiate a Talius::Node object yourself.
	def initialize(raw)
		tokens = raw.split('|', 2)
		
		# if one element
		if tokens.length == 1
			@namespace = nil
			@name = tokens[0]
		elsif tokens.length == 2
			@namespace = tokens[0]
			@name = tokens[1]
		else
			raise 'node-syntax-error: ' + raw.to_s
		end
	end
	#
	# initialize
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# namespace
	#
	
	# The node's namespace. If a namespace is not explicitly given in the
	# selector string then this property is nil.
	#
	#  raw = 'section, mml|dev'
	#  selector = Talius.new(raw)
	#  selector.rules[0].tags['section'].namespace # => nil
	#  selector.rules[1].tags['mml|dev'].namespace # => "mml"
	attr_reader :namespace
	#
	# namespace
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# name
	#
	
	# The name of the tag or attribute. Does not include the namespace.
	#
	#  raw = 'section, mml|dev'
	#  selector = Talius.new(raw)
	#  selector.rules[0].tags['section'].name # => "section"
	#  selector.rules[1].tags['mml|dev'].name # => "dev"
	attr_reader :name
	#
	# name
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# full_name
	#
	
	# Returns the namespace (if there is one) and the name of the node. If there
	# is a namespace then the namespace is followed by |.
	#
	#  raw = 'section, mml|dev'
	#  selector = Talius.new(raw)
	#  selector.rules[0].tags['section'].fullname # => "section"
	#  selector.rules[1].tags['mml|dev'].fullname # => "mml|dev"
	def full_name
		if @namespace
			return @namespace + '|' + @name
		else
			return @name
		end
	end
	#
	# full_name
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# to_s
	#
	
	# Returns the results of +full_name+.
	def to_s
		return full_name
	end
	#
	# to_s
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# to_h
	#
	
	# Returns a hash of the name and namespace (if there is a namespace).
	#
	#  raw = 'section, mml|dev'
	#  selector = Talius.new(raw)
	#  selector.rules[0].tags['section'].to_h # => {"name"=>"section"}
	#  selector.rules[1].tags['mml|dev'].to_h # => {"name"=>"dev", "namespace"=>"mml"}
	def to_h
		rv = {}
		rv['name'] = @name
		
		# namespace
		if @namespace
			rv['namespace'] = @namespace
		end
		
		return rv
	end
	#
	# to_h
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# inspect
	#
	
	# Returns the stringification of the hash returned by +to_h+.
	#  raw = 'section, mml|dev'
	#  selector = Talius.new(raw)
	#  selector.rules[0].tags['section'].inspect # => {"name"=>"section"}
	#  selector.rules[1].tags['mml|dev'].inspect # => {"name"=>"dev", "namespace"=>"mml"}
	def inspect
		return to_h.inspect
	end
	#
	# inspect
	#---------------------------------------------------------------------------
end
#
# Talius::Node
#===============================================================================


#===============================================================================
# Talius::Node::Tag
#

# This class inherits from Talius::Node and adds nothing to it.
class Talius::Node::Tag < Talius::Node
end
#
# Talius::Node::Tag
#===============================================================================


#===============================================================================
# Talius::Node::Att
#

# This class inherits from Talius::Node and adds the value property.
class Talius::Node::Att < Talius::Node
	#---------------------------------------------------------------------------
	# initialize
	#
	
	def initialize(*opts)
		super(*opts)
		@value = nil
	end
	#
	# initialize
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# value
	#
	
	# The value of the attribute rule, if there is one.
	#
	#  raw = 'a[href][rel=parent]'
	#  selector = Talius.new(raw)
	#  rule = selector.rules[0]
	#  rule.atts['href'].value # => nil
	#  rule.atts['rel'].value  # => "parent"
	
	attr_accessor :value
	
	#
	# value
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# to_h
	#
	
	# Returns a hash representation of the object.
	def to_h
		rv = super()
		
		# value
		if @value
			rv['value'] = @value
		end
		
		return rv
	end
	#
	# to_h
	#---------------------------------------------------------------------------
end
#
# Talius::Node::Att
#===============================================================================